import contextlib as _contextlib
import uuid as _uuid
from typing import Generic, Iterator, Optional, Sequence, TypeVar

import sqlalchemy as _sa

from ._compat import SAConnection
from ._exceptions import MessageNotFound
from ._factories import AggregatedStreamFactory
from ._interfaces import (
    MessagePosition,
    MessageProtocol,
    MessageSerializer,
    StoredMessage,
)
from ._storage import Storage

E = TypeVar("E", bound=MessageProtocol)


class MessageStoreReader(Generic[E]):
    """
    Message store reader.
    """

    def __init__(
        self, conn: SAConnection, storage: Storage, serializer: MessageSerializer[E]
    ):
        self._conn = conn
        self._storage = storage
        self._serializer = serializer

    def get_message_by_id(self, message_id: _uuid.UUID) -> StoredMessage[E]:
        """
        Returns a message by ID.

        Args:
            message_id (UUID): Message ID.
        """
        row = self._storage.get_message_by_id(conn=self._conn, message_id=message_id)
        if row:
            message_id, stream, version, message, global_position = row
            return StoredMessage(
                message_id=message_id,
                stream=stream,
                version=version,
                message=self._serializer.deserialize(message),
                global_position=global_position,
            )
        raise MessageNotFound(message_id)

    def get_messages_by_ids(
        self, message_ids: Sequence[_uuid.UUID]
    ) -> Iterator[StoredMessage[E]]:
        """
        Returns multiple messages by IDs.

        Args:
            message_ids (Sequence[UUID]): Message IDs.
        """
        for row in self._storage.get_messages_by_ids(
            conn=self._conn, message_ids=message_ids
        ):
            message_id, stream, version, message, global_position = row
            yield StoredMessage(
                message_id=message_id,
                stream=stream,
                version=version,
                message=self._serializer.deserialize(message),
                global_position=global_position,
            )

    def read(self, stream: str) -> Iterator[StoredMessage[E]]:
        """
        Returns all messages from a stream.

        Args:
            stream (str): Stream name
        """
        for message_id, version, message, global_position in self._storage.read(
            self._conn, stream
        ):
            yield StoredMessage(
                message_id=message_id,
                stream=stream,
                version=version,
                message=self._serializer.deserialize(message),
                global_position=global_position,
            )

    def read_wildcard(self, stream_wildcard: str) -> Iterator[StoredMessage[E]]:
        """
        Returns all messages from streams that match the wildcard.

        Use like syntax to match multiple streams:

        - `stream-%` - match all streams that start with `stream-`
        - `%` - match all streams
        - `%-%` - match all streams that contain `-`

        Args:
            stream_wildcard (str): Stream name wildcard
        """
        for (
            message_id,
            stream,
            version,
            message,
            global_position,
        ) in self._storage.read_wildcard(self._conn, stream_wildcard):
            yield StoredMessage(
                message_id=message_id,
                stream=stream,
                version=version,
                message=self._serializer.deserialize(message),
                global_position=global_position,
            )


class MessageStore(Generic[E]):
    def __init__(
        self,
        name: str,
        engine: _sa.engine.Engine,
        serializer: MessageSerializer[E],
    ):
        """
        Message store.

        Args:
            name (str): A valid python identifier which is used as a prefix for
                the database objects that are created.
            engine (Engine): A SQLAlchemy engine.
            serializer (MessageSerializer): A serializer for the messages.

        Attributes:
            aggregated_stream (AggregatedStreamFactory): A factory for aggregated streams.
        """
        self.engine = engine
        self._storage = Storage(name=name, engine=engine)
        self._serializer = serializer
        self.aggregated_stream = AggregatedStreamFactory(store=self)

    def _get_connection(self) -> SAConnection:
        return self.engine.connect()

    def truncate(self):
        with self._get_connection() as conn:
            self._storage.truncate(conn)
            conn.commit()

    def write(
        self,
        stream: str,
        message: E,
        expected_version: Optional[int] = None,
        conn: Optional[SAConnection] = None,
    ) -> MessagePosition:
        """
        Write a message to the store.

        Optimistic concurrency control can used to ensure that the stream is
        not modified by another process between reading the last message and
        writing the new message. You have to give `expected_version` to use it.
        If the stream has been modified, a `OptimisticConcurrencyError` will be
        raised.

        You can give a connection to use for the write as `conn`. If you don't
        give a connection, a new connection will be created and the write will
        be committed. You can use this to write messages and other data in a
        single transaction. Therefore, if you give a connection, **you have to
        commit the transaction yourself**.

        Args:
            stream (str): The name of the stream to which the message should be written.
            message (MessageProtocol): The message to write.
            expected_version (Optional[int], optional): The expected version of the stream.
            conn (Optional[SAConnection], optional): A database connection.

        Returns:
            MessagePosition: The position of the last message in the stream.
        """
        if conn is None:
            with self._get_connection() as conn:
                result = self._write(
                    conn=conn,
                    stream=stream,
                    message=message,
                    expected_version=expected_version,
                )
                conn.commit()
                return result
        else:
            return self._write(
                conn=conn,
                stream=stream,
                message=message,
                expected_version=expected_version,
            )

    def _write(
        self,
        conn: SAConnection,
        stream: str,
        message: E,
        expected_version: Optional[int] = None,
    ) -> MessagePosition:
        return self._storage.add(
            conn=conn,
            stream=stream,
            expected_version=expected_version,
            message_id=message.get_message_id(),
            message=self._serializer.serialize(message),
        )

    def synchronize(
        self,
        stream: str,
        expected_version: int,
        messages: Sequence[E],
        conn: Optional[SAConnection] = None,
    ) -> MessagePosition:
        """
        Synchronize a stream with a sequence of messages.

        Given a stream and a sequence of messages, this method will write the
        messages to the stream that are not already in it.

        Optimistic concurrency control must used to ensure that the stream is
        not modified by another process between reading the last message and
        writing the new message. You have to give `expected_version`. If the
        stream has been modified, a `OptimisticConcurrencyError` will be raised.

        You can give a connection to use for the write as `conn`. If you don't
        give a connection, a new connection will be created and the write will
        be committed. You can use this to write messages and other data in a
        single transaction. Therefore, if you give a connection, **you have to
        commit the transaction yourself**.

        Args:
            stream (str): The name of the stream to which the message should be written.
            expected_version (int): The expected version of the stream.
            messages (Sequence[MessageProtocol]): The messages that should be in the stream after the synchronization.
            conn (Optional[SAConnection], optional): A database connection.

        Returns:
            MessagePosition: The position of the last message in the stream.
        """
        if conn is None:
            with self._get_connection() as conn:
                result = self._synchronize(
                    conn=conn,
                    stream=stream,
                    expected_version=expected_version,
                    messages=messages,
                )
                conn.commit()
                return result
        else:
            return self._synchronize(
                conn=conn,
                stream=stream,
                expected_version=expected_version,
                messages=messages,
            )

    def _synchronize(
        self,
        conn: SAConnection,
        stream: str,
        expected_version: int,
        messages: Sequence[E],
    ) -> MessagePosition:
        stored_version = self._storage.get_max_version(conn, stream)
        if stored_version is not None:
            stored_ids = list(self._storage.get_message_ids(conn, stream))
            message_ids = [message.get_message_id() for message in messages]
            for stored, given in zip(stored_ids, message_ids):
                if stored != given:
                    raise ValueError("Message ID mismatch")
            messages = messages[len(stored_ids) :]
        if messages:
            result = self._storage.add_all(
                conn,
                stream,
                expected_version,
                [
                    (
                        message.get_message_id(),
                        self._serializer.serialize(message),
                    )
                    for message in messages
                ],
            )
        else:
            result = stored_version
        return result

    def _get_reader(self, conn: SAConnection) -> MessageStoreReader[E]:
        return MessageStoreReader(
            conn=conn,
            storage=self._storage,
            serializer=self._serializer,
        )

    @_contextlib.contextmanager
    def reader(
        self, conn: Optional[SAConnection] = None
    ) -> Iterator[MessageStoreReader[E]]:
        """
        Get a reader for the store.

        You can give a connection to use for the read as `conn`. If you don't
        give a connection, a new connection will be used (and discarded after
        the reader context has been left).

        Example usage:

            with store.reader() as reader:
                message = reader.get_message_id(some_id)

        Args:
            conn (Optional[SAConnection], optional): A database connection.

        Yields:
            MessageStoreReader: A reader for the store.
        """
        if conn:
            yield self._get_reader(conn)
        else:
            with self._get_connection() as conn:
                yield self._get_reader(conn)

    def read(self, stream: str) -> Iterator[StoredMessage[E]]:
        """
        Read all messages from a stream.

        Args:
            stream (str): The name of the stream.

        Returns:
            Iterator[StoredMessage]: An iterator over the messages.
        """
        with self.reader() as reader:
            yield from reader.read(stream)



import contextlib as _contextlib
import datetime as _dt
import uuid as _uuid
from typing import Dict, Generic, Iterator, List, Optional, TypeVar

import sqlalchemy as _sa
from psycopg2.errors import LockNotAvailable
from sqlalchemy_utils import UUIDType as _UUIDType

from ._compat import SAConnection
from ._factories import SubscriptionFactory
from ._interfaces import (
    AggregatedStreamMessage,
    MessagePartitioner,
    MessageProtocol,
    StreamPartitionStatistic,
)
from ._message_store import MessageStore

E = TypeVar("E", bound=MessageProtocol)


class AggregatedStream(Generic[E]):
    def __init__(
        self,
        name: str,
        store: MessageStore[E],
        partitioner: MessagePartitioner[E],
        stream_wildcards: List[str],
        update_batch_size: Optional[int] = None,
    ) -> None:
        """
        AggregatedStream aggregates multiple streams into one (partitioned) stream.

        Read more about aggregated streams under [Data Model](../concepts/data-model.md#aggregated-streams).

        The `update_batch_size` argument can be used to control the batch size of the
        update process. Higher numbers will result in less database roundtrips but
        also in higher memory usage.

        Args:
            name: Stream name, needs to be a valid python identifier
            store: Message store
            partitioner: Message partitioner
            stream_wildcards: List of stream wildcards
            update_batch_size: Batch size for updating the stream, defaults to 100

        Attributes:
            name (str): Stream name
            projector (StreamProjector): Stream projector
            subscription (SubscriptionFactory): Factory to create subscriptions on this stream
        """
        assert name.isidentifier(), "name must be a valid identifier"
        self.name = name
        self.subscription = SubscriptionFactory(self)
        self._store = store
        self._metadata = _sa.MetaData()
        self._table = _sa.Table(
            f"{name}_projected_stream",
            self._metadata,
            _sa.Column("message_id", _UUIDType(), primary_key=True),
            _sa.Column("origin_stream", _sa.String(255), nullable=False),
            _sa.Column("origin_stream_version", _sa.Integer, nullable=False),
            _sa.Column(
                "partition",
                _sa.Integer,
                nullable=False,
                index=True,
            ),
            _sa.Column(
                "position",
                _sa.Integer,
                nullable=False,
            ),
            _sa.Column(
                "message_occurred_at",
                _sa.DateTime,
                nullable=False,
            ),
            _sa.UniqueConstraint(
                "partition",
                "position",
                name=f"uq_{name}_partition_position",
            ),
        )
        self.notification_channel = f"{name}_notifications"
        trigger = _sa.DDL(
            f"""
            CREATE OR REPLACE FUNCTION {name}_stream_notify_message_inserted()
              RETURNS trigger AS $$
            DECLARE
            BEGIN
              PERFORM pg_notify(
                '{self.notification_channel}',
                json_build_object(
                    'message_id', NEW.message_id,
                    'partition', NEW.partition,
                    'position', NEW.position
                )::text);
              RETURN NEW;
            END;
            $$ LANGUAGE plpgsql;

            CREATE TRIGGER {name}_stream_notify_message_inserted
              AFTER INSERT ON {name}_projected_stream
              FOR EACH ROW
              EXECUTE PROCEDURE {name}_stream_notify_message_inserted();
            """
        )
        _sa.event.listen(
            self._table, "after_create", trigger.execute_if(dialect="postgresql")
        )
        self._metadata.create_all(store.engine, checkfirst=True)
        self.projector = StreamProjector(
            stream=self,
            partitioner=partitioner,
            stream_wildcards=stream_wildcards,
            batch_size=update_batch_size,
        )

    def truncate(self, conn: SAConnection):
        """
        Truncate aggregated stream.
        """
        conn.execute(self._table.delete())

    def add(
        self,
        conn: SAConnection,
        message_id: _uuid.UUID,
        stream: str,
        stream_version: int,
        partition: int,
        position: int,
        message_occurred_at: _dt.datetime,
    ) -> None:
        conn.execute(
            self._table.insert().values(
                message_id=message_id,
                origin_stream=stream,
                origin_stream_version=stream_version,
                partition=partition,
                position=position,
                message_occurred_at=message_occurred_at,
            )
        )

    def read(self, partition: int) -> Iterator[AggregatedStreamMessage]:
        """
        Read all messages from a partition of the aggregated stream.

        Args:
            partition: Partition number
        """
        with self._connection() as conn:
            for row in conn.execute(
                _sa.select(self._table.c.message_id, self._table.c.position)
                .where(self._table.c.partition == partition)
                .order_by(self._table.c.position)
            ):
                yield AggregatedStreamMessage(
                    message_id=row.message_id,
                    position=row.position,
                    partition=partition,
                )

    def read_slice(
        self, partition: int, start: int, count: int
    ) -> Iterator[AggregatedStreamMessage]:
        """
        Read a slice of messages from a partition of the aggregated stream.

        Args:
            partition: Partition number
            start: Start position
            count: Number of messages to read
        """
        with self._connection() as conn:
            for row in conn.execute(
                _sa.select(self._table.c.message_id, self._table.c.position)
                .where(
                    _sa.and_(
                        self._table.c.partition == partition,
                        self._table.c.position >= start,
                    )
                )
                .order_by(self._table.c.position)
                .limit(count)
            ):
                yield AggregatedStreamMessage(
                    message_id=row.message_id,
                    position=row.position,
                    partition=partition,
                )

    @_contextlib.contextmanager
    def _connection(self):
        conn = self._store.engine.connect()
        try:
            yield conn
        finally:
            conn.close()

    def get_partition_statistics(
        self,
        position_limits: Dict[int, int] = None,
        result_limit: Optional[int] = None,
    ) -> Iterator[StreamPartitionStatistic]:
        """
        Get partition statistics for deciding which partitions to read from. This
        is used by subscriptions.
        """
        with self._connection() as conn:
            position_limits = position_limits or {-1: -1}
            tbl = self._table.alias()
            next_messages_tbl = (
                _sa.select(
                    tbl.c.partition,
                    _sa.func.min(tbl.c.position).label("min_position"),
                )
                .where(
                    _sa.or_(
                        *[
                            _sa.and_(
                                tbl.c.partition == partition, tbl.c.position > limit
                            )
                            for partition, limit in position_limits.items()
                        ],
                        _sa.not_(tbl.c.partition.in_(list(position_limits))),
                    )
                )
                .group_by(tbl.c.partition)
                .cte()
            )

            qry = (
                _sa.select(tbl)
                .select_from(
                    next_messages_tbl.join(
                        tbl,
                        _sa.and_(
                            tbl.c.partition == next_messages_tbl.c.partition,
                            tbl.c.position == next_messages_tbl.c.min_position,
                        ),
                    )
                )
                .order_by(tbl.c.message_occurred_at)
                .limit(result_limit)
            )
            result = conn.execute(qry)
            for row in result.fetchall():
                yield StreamPartitionStatistic(
                    partition_number=row.partition,
                    next_message_id=row.message_id,
                    next_message_position=row.position,
                    next_message_occurred_at=row.message_occurred_at,
                )
            result.close()
            del result


class _AlreadyUpdating(RuntimeError):
    pass


class StreamProjector(Generic[E]):
    def __init__(
        self,
        stream: AggregatedStream[E],
        partitioner: MessagePartitioner[E],
        stream_wildcards: List[str],
        batch_size: Optional[int] = None,
    ):
        """
        Stream projector is responsible for updating an aggregated stream.

        The update process is locked to prevent concurrent updates. Thus, it is
        fine to run the projector in multiple processes.

        Implements: [RunOnNotification][depeche_db.RunOnNotification]
        """
        self.stream = stream
        self.stream_wildcards = stream_wildcards
        self.partitioner = partitioner
        self.batch_size = batch_size or 100

    @property
    def notification_channel(self) -> str:
        """
        Returns the notification channel name for this projector.
        """
        return self.stream._store._storage.notification_channel

    def run(self):
        """
        Runs the projector once.
        """
        try:
            self.update_full()
        except _AlreadyUpdating:
            pass

    def stop(self):
        """
        No-Op on this class.
        """
        pass

    def update_full(self) -> int:
        """
        Updates the projection from the last known position to the current position.
        """
        result = 0
        with self.stream._store.engine.connect() as conn:
            cutoff = self.stream._store._storage.get_global_position(conn)
            try:
                conn.execute(
                    _sa.text(
                        f"LOCK TABLE {self.stream._table.name} IN EXCLUSIVE MODE NOWAIT"
                    )
                )
            except _sa.exc.OperationalError as exc:
                if isinstance(exc.orig, LockNotAvailable):
                    raise _AlreadyUpdating(
                        "Cannot update stream projection, because another process is already updating it."
                    )
                raise
            while True:
                batch_num = self._update_batch(conn, cutoff)
                if batch_num == 0:
                    break
                result += batch_num
            conn.commit()
        return result

    def _update_batch(self, conn, cutoff: Optional[int] = None) -> int:
        tbl = self.stream._table.alias()
        message_table = self.stream._store._storage.message_table
        last_seen = (
            _sa.select(
                tbl.c.origin_stream,
                _sa.func.max(tbl.c.origin_stream_version).label("max_version"),
            )
            .group_by(tbl.c.origin_stream)
            .cte()
        )
        cutoff_cond = []
        if cutoff is not None:
            cutoff_cond = [message_table.c.global_position <= cutoff]
        q = (
            _sa.select(
                message_table.c.message_id,
                message_table.c.stream,
                message_table.c.version,
            )
            .select_from(
                message_table.join(
                    last_seen,
                    message_table.c.stream == last_seen.c.origin_stream,
                    isouter=True,
                )
            )
            .where(
                _sa.and_(
                    _sa.or_(
                        *[
                            message_table.c.stream.like(wildcard)
                            for wildcard in self.stream_wildcards
                        ]
                    ),
                    _sa.or_(
                        message_table.c.version > last_seen.c.max_version,
                        last_seen.c.max_version.is_(None),
                    ),
                    *cutoff_cond,
                )
            )
            .order_by(message_table.c.global_position)
            .limit(self.batch_size)
        )
        messages = list(conn.execute(q))
        self._add(conn, messages)
        return len(messages)

    def _add(self, conn, messages):
        positions = {
            row.partition: row.max_position
            for row in conn.execute(
                _sa.select(
                    self.stream._table.c.partition,
                    _sa.func.max(self.stream._table.c.position).label("max_position"),
                ).group_by(self.stream._table.c.partition)
            )
        }

        with self.stream._store.reader(conn) as reader:
            stored_messages = {
                message.message_id: message
                for message in reader.get_messages_by_ids(
                    [message_id for message_id, *_ in messages]
                )
            }

        for message_id, stream, version in messages:
            message = stored_messages[message_id]
            partition = self.partitioner.get_partition(message)
            position = positions.get(partition, -1) + 1
            self.stream.add(
                conn=conn,
                message_id=message_id,
                stream=stream,
                stream_version=version,
                partition=partition,
                position=position,
                message_occurred_at=message.message.get_message_time(),
            )
            positions[partition] = position



import collections as _collections
import queue as _queue
import signal as _signal
import threading as _threading
import time as _time
from typing import Dict, List, Optional

from ._interfaces import RunOnNotification
from .tools import PgNotificationListener


class Executor:
    """
    Executor is a class that runs handlers on notifications.

    Typical usage:

        executor = Executor(db_dsn="postgresql://localhost:5432/mydb")
        executor.register(MyRunOnNotification())
        executor.run()   # this will stop on SIGTERM or SIGINT

    Args:
        db_dsn: DSN for the PostgreSQL database
    """

    listener: Optional[PgNotificationListener] = None

    def __init__(self, db_dsn: str):
        self._db_dsn = db_dsn
        self.channel_register: Dict[
            str, List[RunOnNotification]
        ] = _collections.defaultdict(list)
        self.stimulation_interval = 0.5
        self.keep_running = True
        self.handler_queue = UniqueQueue()
        self.stimulator_thread = _threading.Thread(target=self._stimulate, daemon=True)
        self.handler_thread = _threading.Thread(target=self._run_handlers, daemon=True)
        self.listener = None
        _signal.signal(_signal.SIGINT, lambda *_: self._stop())
        _signal.signal(_signal.SIGTERM, lambda *_: self._stop())

    def register(self, handler: RunOnNotification):
        """
        Registers a handler to be run on notifications.

        Args:
            handler: Handler to register
        """
        self.channel_register[handler.notification_channel].append(handler)
        return handler

    def _stop(self):
        self.keep_running = False
        if self.listener is not None:
            self.listener.stop()

        for handlers in self.channel_register.values():
            for handler in handlers:
                handler.stop()

    def run(self):
        """Runs the executor."""
        self.listener = PgNotificationListener(
            dsn=self._db_dsn,
            channels=list(self.channel_register),
            ignore_payload=True,
        )

        self.handler_thread.start()
        self.stimulator_thread.start()
        self.listener.start()

        for notification in self.listener.messages():
            for handler in self.channel_register[notification.channel]:
                self.handler_queue.put(handler.run)

        self.handler_thread.join()
        self.stimulator_thread.join()

    def _run_handlers(self):
        while self.keep_running:
            try:
                handler = self.handler_queue.get(timeout=0.5)
                try:
                    handler()
                except Exception:
                    self._stop()
                    raise
            except _queue.Empty:
                pass

    def _stimulate(self):
        while self.keep_running:
            for handlers in self.channel_register.values():
                for handler in handlers:
                    self.handler_queue.put(handler.run)
            _time.sleep(self.stimulation_interval)


class UniqueQueue(_queue.Queue):
    def _put(self, item):
        if item not in self.queue:
            self.queue.append(item)



import dataclasses as _dc
import datetime as _dt
import enum as _enum
import uuid as _uuid
from typing import (
    Callable,
    Dict,
    Generic,
    Iterator,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    no_type_check,
    runtime_checkable,
)


@runtime_checkable
class MessageProtocol(Protocol):
    """
    Message protocol is a base class for all messages that are used in the system.
    """

    def get_message_id(self) -> _uuid.UUID:
        """
        Returns message ID
        """
        ...

    def get_message_time(self) -> _dt.datetime:
        """
        Returns message time
        """
        ...


E = TypeVar("E", bound=MessageProtocol)
M = TypeVar("M")


@_dc.dataclass(frozen=True)
class StoredMessage(Generic[E]):
    """
    Stored message is a message that is stored in the stream.

    Attributes:
        message_id: Message ID
        stream: Stream name
        version: Message version
        message: Message (`E` subtype of `MessageProtocol`)
        global_position: Global position
    """

    message_id: _uuid.UUID
    stream: str
    version: int
    message: E
    global_position: int


@_dc.dataclass(frozen=True)
class SubscriptionMessage(Generic[E]):
    """
    Subscription message is a message that is received from the subscription.

    Attributes:
        partition: Partition number
        position: Position in the partition
        stored_message: Stored message (`E` subtype of `MessageProtocol`)
    """

    partition: int
    position: int
    stored_message: StoredMessage[E]


@_dc.dataclass(frozen=True)
class MessagePosition:
    """
    Message position is a position of the message in the stream.

    Attributes:
        stream: Stream name
        version: Message version
        global_position: Global position
    """

    stream: str
    version: int
    global_position: int


@_dc.dataclass
class StreamPartitionStatistic:
    partition_number: int
    next_message_id: _uuid.UUID
    next_message_position: int
    next_message_occurred_at: _dt.datetime


@_dc.dataclass
class AggregatedStreamMessage:
    """
    Aggregated stream message is a message that can be read from an aggregated stream.

    Attributes:
        message_id: Message ID
        partition: Partition number
        position: Position in the partition
    """

    partition: int
    position: int
    message_id: _uuid.UUID


@_dc.dataclass
class SubscriptionState:
    """
    Subscription state is a state of the subscription.

    Attributes:
        positions: Mapping of partition number to the position in the partition
    """

    positions: Dict[int, int]


class MessageSerializer(Protocol, Generic[M]):
    """
    Message serializer is a protocol that is used to serialize and deserialize messages.

    The following must be true for any serializer:

    - `deserialize(serialize(message)) == message`
    - `type(deserialize(serialize(message))) is type(message)`
    - `serialize(deserialize(data)) == data`
    """

    def serialize(self, message: M) -> dict:
        """
        Serializes message to a dictionary. The dictionary must be JSON serializable.
        """
        raise NotImplementedError()

    def deserialize(self, message: dict) -> M:
        """
        Deserializes message from a dictionary.
        """
        raise NotImplementedError()


class MessagePartitioner(Protocol, Generic[E]):
    """
    Message partitioner is a protocol that is used to determine partition number for a message.
    """

    def get_partition(self, message: StoredMessage[E]) -> int:
        """
        Returns partition number for a message. The partition number must be a
        positive integer. The partition number must be deterministic for a given message.
        """
        raise NotImplementedError


class LockProvider(Protocol):
    """
    Lock provider is a protocol that is used to lock and unlock resources.
    """

    def lock(self, name: str) -> bool:
        """
        Locks resource with a given name. Returns `True` if the resource was locked.
        This method must not block!
        """
        raise NotImplementedError

    def unlock(self, name: str):
        """
        Unlocks resource with a given name.
        """
        raise NotImplementedError


class SubscriptionStateProvider(Protocol):
    """
    Subscription state provider is a protocol that is used to store and read subscription state.
    """

    def store(self, subscription_name: str, partition: int, position: int):
        """
        Stores subscription state for a given partition.
        """
        raise NotImplementedError

    def read(self, subscription_name: str) -> SubscriptionState:
        """
        Reads subscription state.

        Returns:
            Subscription state
        """
        raise NotImplementedError


class CallMiddleware(Generic[E]):
    """
    Call middleware is a protocol that is used to wrap a call to a handler.

    Typical implementation:

        class MyCallMiddleware(CallMiddleware):
            def __init__(self, some_dependency):
                self.some_dependency = some_dependency

            def call(self, handler, message):
                # or use a DI container here
                handler(message, some_dependency=self.some_dependency)
    """

    def call(
        self,
        handler: Callable,
        message: Union[SubscriptionMessage[E], StoredMessage[E], E],
    ):
        """
        Calls a handler with a given message.

        The type of the message depends on the type annotation of the handler function.
        See [MessageHandlerRegister][depeche_db.MessageHandlerRegister] for more details.

        Args:
            handler: Handler
            message: Message to be passed to the handler
        """
        raise NotImplementedError


class RunOnNotification(Protocol):
    """
    Run on notification is a protocol that allows objects to be run when a
    notification is received on a channel. Objects that implement this protocol
    can be registered with a [Executor][depeche_db.Executor] object.

    Implemented by:
        - [SubscriptionRunner][depeche_db.SubscriptionRunner]
        - [StreamProjector][depeche_db.StreamProjector]
    """

    @property
    def notification_channel(self) -> str:
        """
        Returns notification channel name.
        """
        raise NotImplementedError

    def run(self):
        """
        Runs the object. This method needs to return when a chunk of work has been
        done.
        """
        raise NotImplementedError

    def stop(self):
        """
        If the object's `run` method has a loop, this method can be used to
        exit the loop earlier.
        Will be called in a separate thread.
        """
        raise NotImplementedError


class ErrorAction(_enum.Enum):
    """
    Error action is an action that is taken when an error occurs during message processing.

    Attributes:
        IGNORE: Ignore the error and continue processing.
        EXIT: Exit processing.
    """

    IGNORE = "ignore"
    EXIT = "exit"


class SubscriptionErrorHandler(Generic[E]):
    """
    Subscription error handler is a protocol that is used to handle errors that occur.
    """

    def handle_error(
        self, error: Exception, message: SubscriptionMessage[E]
    ) -> ErrorAction:
        """
        Handles an error that occurred during message processing.

        Args:
            error: Error
            message: Message that was being processed when the error occurred

        Returns:
            Action to be taken
        """
        raise NotImplementedError


@_dc.dataclass
class HandlerDescriptor(Generic[E]):
    handler: Callable
    pass_subscription_message: bool
    pass_stored_message: bool
    requires_middleware: bool

    @no_type_check
    def adapt_message_type(
        self, message: Union[SubscriptionMessage[E], StoredMessage[E], E]
    ) -> Union[SubscriptionMessage[E], StoredMessage[E], E]:
        if isinstance(message, SubscriptionMessage):
            if self.pass_subscription_message:
                return message
            elif self.pass_stored_message:
                return message.stored_message
            else:
                return message.stored_message.message
        elif isinstance(message, StoredMessage):
            if self.pass_subscription_message:
                raise ValueError(
                    "SubscriptionMessage was requested, but StoredMessage was provided"
                )
            elif self.pass_stored_message:
                return message.stored_message
            else:
                return message.stored_message.message
        else:
            if self.pass_subscription_message or self.pass_stored_message:
                raise ValueError(
                    "SubscriptionMessage or StoredMessage was requested, but plain message was provided"
                )
            else:
                return message


class MessageHandlerRegisterProtocol(Protocol, Generic[E]):
    """
    Message handler register protocol is used by runners to find handlers for messages.

    Implemented by:
        - [MessageHandlerRegister][depeche_db.MessageHandlerRegister]
        - [MessageHandler][depeche_db.MessageHandler]
    """

    def get_all_handlers(self) -> Iterator[HandlerDescriptor[E]]:
        """
        Returns all registered handlers.
        """
        raise NotImplementedError

    def get_handler(self, message_type: Type[E]) -> Optional[HandlerDescriptor[E]]:
        """
        Returns a handler for a given message type.
        """
        raise NotImplementedError



import datetime as _dt
import decimal as _decimal
import enum as _enum
import uuid as _uuid
from typing import Any, Optional, Type, TypeVar, no_type_check

from depeche_db._compat import get_union_members, issubclass_with_union

from .._interfaces import MessageSerializer

T = TypeVar("T")


class PydanticMessageSerializer(MessageSerializer[T]):
    # the real type would be: (self, message_type: Type[T])
    # but this is not supported by mypy (yet)
    def __init__(
        self, message_type: Any, aliases: Optional[dict[str, Type]] = None
    ) -> None:
        import pydantic

        version = pydantic.VERSION.split(".")[0]
        self.pydantic_v2 = version == "2"

        self.message_type: type[T] = message_type
        self.message_types: dict[str, Type] = {}
        for member in get_union_members(message_type):
            if member.__name__ in self.message_types:
                raise ValueError(
                    f"Duplicate class name {member.__name__} in union {message_type}"
                )
            self.message_types[member.__name__] = member

        if aliases is not None:
            for alias, member in aliases.items():
                if alias in self.message_types:
                    raise ValueError(
                        f"Alias {alias} conflicts with existing class name in union {message_type}"
                    )
                if not issubclass_with_union(member, self.message_type):
                    raise ValueError(
                        f"Type {member} is not a subclass of {self.message_type}"
                    )
                self.message_types[alias] = member

    def serialize(self, message: T) -> dict:
        typename = message.__class__.__name__
        if typename not in self.message_types:
            raise TypeError(
                f"Unknown message type {typename} (expected one of {self.message_types})"
            )
        result: dict = self._serialize(message)
        result["__typename__"] = typename
        return result

    @no_type_check
    def _serialize(self, message: T) -> dict:
        if self.pydantic_v2:
            return message.model_dump(mode="json")
        return make_jsonable(message.dict())

    @no_type_check
    def deserialize(self, message: dict) -> T:
        import pydantic as _pydantic

        typename = message.pop("__typename__")
        if typename not in self.message_types:
            raise TypeError(
                f"Unknown message type {typename} (expected one of {self.message_types})"
            )
        message_type = self.message_types[typename]

        if self.pydantic_v2:
            return _pydantic.TypeAdapter(message_type).validate_python(message)
        return _pydantic.parse_obj_as(message_type, message)


def make_jsonable(obj):
    if isinstance(obj, list):
        return [make_jsonable(elem) for elem in obj]
    if isinstance(obj, dict):
        return {key: make_jsonable(value) for key, value in obj.items()}
    if isinstance(obj, _uuid.UUID):
        return str(obj)
    if isinstance(obj, _decimal.Decimal):
        return str(obj)
    if isinstance(obj, _dt.datetime):
        return obj.isoformat()
    if isinstance(obj, _dt.date):
        return obj.isoformat()
    if isinstance(obj, _enum.Enum):
        return obj.value
    if obj is None or isinstance(obj, (str, int, float, bool)):
        return obj
    raise NotImplementedError(f"Cannot make object of type {type(obj)} jsonable")



import logging as _logging
from typing import (
    Callable,
    Generic,
    Iterator,
    Optional,
    TypeVar,
    Union,
)

from . import tools as _tools
from ._aggregated_stream import AggregatedStream
from ._interfaces import (
    CallMiddleware,
    ErrorAction,
    LockProvider,
    MessageHandlerRegisterProtocol,
    MessageProtocol,
    StoredMessage,
    SubscriptionErrorHandler,
    SubscriptionMessage,
    SubscriptionStateProvider,
)

E = TypeVar("E", bound=MessageProtocol)

DEPECHE_LOGGER = _logging.getLogger("depeche_db")


class ExitSubscriptionErrorHandler(SubscriptionErrorHandler):
    """
    Exit the subscription on error
    """

    def handle_error(
        self, error: Exception, message: SubscriptionMessage[E]
    ) -> ErrorAction:
        return ErrorAction.EXIT


class LogAndIgnoreSubscriptionErrorHandler(SubscriptionErrorHandler):
    """
    Log the error and ignore the message
    """

    def __init__(self, subscription_name: str):
        self._logger = _logging.getLogger(
            f"depeche_db.subscription.{subscription_name}"
        )

    def handle_error(
        self, error: Exception, message: SubscriptionMessage[E]
    ) -> ErrorAction:
        self._logger.exception(
            "Error while handling message {message.stored_message.message_id}:{message.stored_message.message.__class__.__name__}"
        )
        return ErrorAction.IGNORE


class Subscription(Generic[E]):
    def __init__(
        self,
        name: str,
        stream: AggregatedStream[E],
        message_handler: "SubscriptionMessageHandler[E]",
        batch_size: Optional[int] = None,
        state_provider: Optional[SubscriptionStateProvider] = None,
        lock_provider: Optional[LockProvider] = None,
        # TODO start at time
        # TODO start at "next message"
    ):
        """
        A subscription is a way to read messages from an aggregated stream.

        Read more about the subscription in the [concepts section](../concepts/subscriptions.md).

        Args:
            name: Name of the subscription, needs to be a valid python identifier
            stream: Stream to read from
            message_handler: Handler for the messages
            batch_size: Number of messages to read at once, defaults to 10, read more [here][depeche_db.SubscriptionRunner]
            state_provider: Provider for the subscription state, defaults to a PostgreSQL provider
            lock_provider: Provider for the locks, defaults to a PostgreSQL provider
        """
        assert name.isidentifier(), "Group name must be a valid identifier"
        self.name = name
        self._stream = stream
        self._lock_provider = lock_provider or _tools.DbLockProvider(
            name, self._stream._store.engine
        )
        self._state_provider = state_provider or _tools.DbSubscriptionStateProvider(
            name, self._stream._store.engine
        )
        self.runner = SubscriptionRunner(
            subscription=self,
            message_handler=message_handler,
            batch_size=batch_size,
        )

    def get_next_messages(self, count: int) -> Iterator[SubscriptionMessage[E]]:
        state = self._state_provider.read(self.name)
        statistics = list(
            self._stream.get_partition_statistics(
                position_limits=state.positions, result_limit=10
            )
        )
        for statistic in statistics:
            lock_key = f"subscription-{self.name}-{statistic.partition_number}"
            if not self._lock_provider.lock(lock_key):
                continue
            # now we have the lock, we need to check if the position is still valid
            # if not, we need to release the lock and try the next partition
            state = self._state_provider.read(self.name)
            if state.positions.get(statistic.partition_number, -1) != (
                statistic.next_message_position - 1
            ):
                self._lock_provider.unlock(lock_key)
                continue
            try:
                with self._stream._store.reader() as reader:
                    message_pointers = list(
                        self._stream.read_slice(
                            partition=statistic.partition_number,
                            start=statistic.next_message_position,
                            count=count,
                        )
                    )
                    stored_messages = {
                        message.message_id: message
                        for message in reader.get_messages_by_ids(
                            [pointer.message_id for pointer in message_pointers]
                        )
                    }

                    for pointer in message_pointers:
                        yield SubscriptionMessage(
                            partition=pointer.partition,
                            position=pointer.position,
                            stored_message=stored_messages[pointer.message_id],
                        )
                        self._ack(
                            partition=pointer.partition,
                            position=pointer.position,
                        )
                break
            finally:
                self._lock_provider.unlock(lock_key)

    def _ack(self, partition: int, position: int):
        state = self._state_provider.read(self.name)
        assert (
            state.positions.get(partition, -1) == position - 1
        ), f"{partition} should have {position - 1} as last position, but has {state.positions.get(partition, -1)}"
        self._state_provider.store(
            subscription_name=self.name, partition=partition, position=position
        )


class SubscriptionMessageHandler(Generic[E]):
    def __init__(
        self,
        handler_register: MessageHandlerRegisterProtocol[E],
        error_handler: Optional[SubscriptionErrorHandler] = None,
        call_middleware: Optional[CallMiddleware] = None,
    ):
        """
        Handles messages

        Args:
            handler_register: The handler register to use
            error_handler: A handler for errors raised by the handlers, defaults to handler that will exit the subscription
            call_middleware: The middleware to call before calling the handler
        """
        self._register = handler_register
        self._error_handler = error_handler or ExitSubscriptionErrorHandler()
        self._call_middleware = call_middleware

        if not self._call_middleware and any(
            handler.requires_middleware for handler in self._register.get_all_handlers()
        ):
            raise ValueError(
                "If handler has more than one parameter, a call_middleware must be provided"
            )

    def handle(self, message: SubscriptionMessage):
        handler = self._register.get_handler(type(message.stored_message.message))
        if handler:
            try:
                self._exec(handler.handler, handler.adapt_message_type(message))
            except Exception as error:
                error_handling_result = self._error_handler.handle_error(
                    error=error, message=message
                )
                if error_handling_result == ErrorAction.EXIT:
                    raise

    def _exec(
        self,
        handler: Callable[..., None],
        message: Union[SubscriptionMessage[E], StoredMessage[E], E],
    ) -> None:
        if self._call_middleware:
            self._call_middleware.call(handler, message)
        else:
            handler(message)


class SubscriptionRunner(Generic[E]):
    @classmethod
    def create(
        cls,
        subscription: Subscription[E],
        handlers: MessageHandlerRegisterProtocol[E],
        call_middleware: Optional[CallMiddleware[E]] = None,
        error_handler: Optional[SubscriptionErrorHandler[E]] = None,
    ) -> "SubscriptionRunner[E]":
        return cls(
            subscription=subscription,
            message_handler=SubscriptionMessageHandler(
                handler_register=handlers,
                call_middleware=call_middleware,
                error_handler=error_handler,
            ),
        )

    def __init__(
        self,
        subscription: Subscription[E],
        message_handler: SubscriptionMessageHandler,
        batch_size: Optional[int] = None,
    ):
        """
        Handles messages from a subscription using a handler

        The `batch_size` argument controls how many messages to handle in each
        batch. If not provided, the default is 10. A larger batch size will
        result less round trips to the database, but will also make it more
        likely that messages from _different partitions_ will be processed out of
        the order defined by their `global_position` on the message store.

        A batch size of 1 will ensure that messages are processed in order
        regarding to their `global_position`.
        Messages in the same partition will always be processed in order.

        Implements: [RunOnNotification][depeche_db.RunOnNotification]

        Args:
            subscription: The subscription to handle
            message_handler: The handler to use
            batch_size: The number of messages to handle in each batch, defaults to 10
        """
        self._subscription = subscription
        self._batch_size = batch_size or 10
        self._keep_running = True
        self._handler = message_handler

    @property
    def notification_channel(self) -> str:
        return self._subscription._stream.notification_channel

    def run(self):
        self.run_once()

    def stop(self):
        self._keep_running = False

    def run_once(self):
        while self._keep_running:
            n = 0
            for message in self._subscription.get_next_messages(count=self._batch_size):
                n += 1
                self.handle(message)
            if n == 0:
                break

    def handle(self, message: SubscriptionMessage):
        self._handler.handle(message)



from typing import (
    TYPE_CHECKING,
    Generic,
    List,
    Optional,
    TypeVar,
)

from ._interfaces import (
    CallMiddleware,
    LockProvider,
    MessageHandlerRegisterProtocol,
    MessagePartitioner,
    MessageProtocol,
    SubscriptionErrorHandler,
    SubscriptionStateProvider,
)

if TYPE_CHECKING:
    from ._aggregated_stream import AggregatedStream
    from ._message_store import MessageStore
    from ._subscription import Subscription

E = TypeVar("E", bound=MessageProtocol)


class AggregatedStreamFactory(Generic[E]):
    def __init__(self, store: "MessageStore[E]"):
        """
        This factory is accessible on the message store:

            store = MessageStore(...)
            stream = store.aggregated_stream(
                name="stream_name",
                partitioner=...,
                stream_wildcards=["stream_%"]
            )
        """
        self._store = store

    def __call__(
        self,
        name: str,
        partitioner: "MessagePartitioner[E]",
        stream_wildcards: List[str],
        update_batch_size: Optional[int] = None,
    ) -> "AggregatedStream[E]":
        """
        Create an aggregated stream.

        Args:
            name: The name of the stream, needs to be a valid python identifier
            partitioner: A partitioner for the stream
            stream_wildcards: A list of stream wildcards
            update_batch_size: The batch size for updating the stream
        """
        from ._aggregated_stream import AggregatedStream

        return AggregatedStream(
            store=self._store,
            name=name,
            partitioner=partitioner,
            stream_wildcards=stream_wildcards,
        )


class SubscriptionFactory(Generic[E]):
    def __init__(self, stream: "AggregatedStream[E]"):
        """
        This factory is accessible on the aggregated stream:

            stream : AggregatedStream = ...
            subscription = stream.subscription(
                name="subscription_name",
                handlers=...,
                call_middleware=...,
                error_handler=...,
                state_provider=...,
                lock_provider=...,
            )
        """
        self._stream = stream

    def __call__(
        self,
        name: str,
        handlers: Optional[MessageHandlerRegisterProtocol[E]] = None,
        batch_size: Optional[int] = None,
        call_middleware: Optional[CallMiddleware] = None,
        error_handler: Optional[SubscriptionErrorHandler] = None,
        state_provider: Optional[SubscriptionStateProvider] = None,
        lock_provider: Optional[LockProvider] = None,
    ) -> "Subscription[E]":
        """
        Create a subscription.

        Args:
            name: The name of the subscription, needs to be a valid python identifier
            handlers: Handlers to be called when a message is received, defaults to an empty register
            batch_size: Number of messages to read at once, defaults to 10, read more [here][depeche_db.SubscriptionRunner]
            call_middleware: A middleware to customize the call to the handlers
            error_handler: A handler for errors raised by the handlers, defaults to handler that will exit the subscription
            state_provider: Provider for the subscription state, defaults to a PostgreSQL provider
            lock_provider: Provider for the locks, defaults to a PostgreSQL provider
        """
        from ._message_handler import MessageHandlerRegister
        from ._subscription import Subscription, SubscriptionMessageHandler

        if handlers is None:
            # allow constructing a subscription without handlers
            handlers = MessageHandlerRegister()

        return Subscription(
            name=name,
            stream=self._stream,
            message_handler=SubscriptionMessageHandler(
                handler_register=handlers,
                call_middleware=call_middleware,
                error_handler=error_handler,
            ),
            state_provider=state_provider,
            lock_provider=lock_provider,
        )



import dataclasses as _dc
import json
import logging
import queue
import select
import threading
from typing import Iterator, Sequence

import psycopg2

logger = logging.getLogger(__name__)


@_dc.dataclass
class PgNotification:
    channel: str
    payload: dict


class PgNotificationListener:
    def __init__(
        self,
        dsn: str,
        channels: Sequence[str],
        select_timeout: float = 3.0,
        ignore_payload: bool = False,
    ):
        self.dsn = dsn
        self.channels = channels
        self._keep_running = True
        self._thread = threading.Thread(target=self._loop)
        self._queue = queue.Queue[PgNotification]()
        self._ignore_payload = ignore_payload
        self._select_timeout = select_timeout
        self._queue_timeout = select_timeout / 2

    def messages(self) -> Iterator[PgNotification]:
        while self._keep_running:
            try:
                yield self._queue.get(block=True, timeout=self._queue_timeout)
            except queue.Empty:
                pass

    def start(self):
        self._thread.start()

    def stop(self):
        self._keep_running = False
        self._thread.join()

    def _loop(self):
        conn = psycopg2.connect(self.dsn)

        try:
            curs = conn.cursor()
            for channel in self.channels:
                curs.execute(f"LISTEN {channel};")
            curs.execute("commit;")

            while self._keep_running:
                if select.select([conn], [], [], self._select_timeout) == ([], [], []):
                    pass
                else:
                    conn.poll()
                    while conn.notifies:
                        notification = conn.notifies.pop(0)
                        try:
                            if self._ignore_payload:
                                payload = {}
                            else:
                                payload = json.loads(notification.payload)
                            self._queue.put(
                                PgNotification(notification.channel, payload)
                            )
                        except Exception:
                            logger.exception(
                                f"Error processing notification payload: {notification}"
                            )

        finally:
            conn.close()



import abc as _abc
from typing import Generic, TypeVar

from depeche_db import MessageProtocol

ID = TypeVar("ID")
E = TypeVar("E", bound=MessageProtocol)


class EventSourcedAggregateRoot(_abc.ABC, Generic[ID, E]):
    def __init__(self):
        self._events: list[E] = []
        self._version = 0

    @property
    def events(self) -> list[E]:
        return list(self._events)

    @_abc.abstractmethod
    def get_id(self) -> ID:
        raise NotImplementedError

    def _add_event(self, event: E) -> None:
        self._version += 1
        self._events.append(event)

    def apply(self, event: E) -> None:
        self._apply(event)
        self._check_invariants()
        self._add_event(event)

    @_abc.abstractmethod
    def _apply(self, event: E) -> None:
        raise NotImplementedError

    def _check_invariants(self) -> None:
        pass

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, EventSourcedAggregateRoot):
            raise NotImplementedError()
        return self.get_id() == other.get_id() and self._version == other._version



import sqlalchemy as _sa

from .._interfaces import SubscriptionState


class DbSubscriptionStateProvider:
    def __init__(self, name: str, engine: _sa.engine.Engine):
        assert name.isidentifier(), "Name must be a valid identifier"
        self.name = name
        self._engine = engine

        self.metadata = _sa.MetaData()
        self.state_table = _sa.Table(
            f"{name}_subscription_state",
            self.metadata,
            _sa.Column("subscription_name", _sa.String, primary_key=True),
            _sa.Column("partition", _sa.Integer, primary_key=True),
            _sa.Column("position", _sa.Integer, nullable=False),
        )
        self.metadata.create_all(self._engine)

    def store(self, subscription_name: str, partition: int, position: int):
        from sqlalchemy.dialects.postgresql import insert

        with self._engine.connect() as conn:
            conn.execute(
                insert(self.state_table)
                .values(
                    subscription_name=subscription_name,
                    partition=partition,
                    position=position,
                )
                .on_conflict_do_update(
                    index_elements=[
                        self.state_table.c.subscription_name,
                        self.state_table.c.partition,
                    ],
                    set_={
                        self.state_table.c.position: position,
                    },
                )
            )
            conn.commit()

    def read(self, subscription_name: str) -> SubscriptionState:
        with self._engine.connect() as conn:
            return SubscriptionState(
                {
                    row.partition: row.position
                    for row in conn.execute(
                        _sa.select(
                            self.state_table.c.partition,
                            self.state_table.c.position,
                        ).where(
                            self.state_table.c.subscription_name == subscription_name
                        )
                    )
                }
            )
